import sqlite3
import unittest

from dependency_injector import containers

# Runtime import
import os
_TOP_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        '../',
    )),
)
_SAMPLES_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        '../samples/',
    )),
)
import sys
sys.path.append(_SAMPLES_DIR)

from schemasample.services import UserService, AuthService, PhotoService


class TestSchemaSingleContainer(unittest.TestCase):

    def test(self):
        container = containers.DynamicContainer()
        container.from_yaml_schema(f'{_SAMPLES_DIR}/schemasample/container-single.yml')
        container.config.from_dict({
            'database': {
                'dsn': ':memory:',
            },
            'aws': {
                'access_key_id': 'KEY',
                'secret_access_key': 'SECRET',
            },
            'auth': {
                'token_ttl': 3600,
            },
        })

        # User service
        user_service1 = container.user_service()
        user_service2 = container.user_service()
        self.assertIsInstance(user_service1, UserService)
        self.assertIsInstance(user_service2, UserService)
        self.assertIsNot(user_service1, user_service2)

        self.assertIsInstance(user_service1.db, sqlite3.Connection)
        self.assertIsInstance(user_service2.db, sqlite3.Connection)
        self.assertIs(user_service1.db, user_service2.db)

        # Auth service
        auth_service1 = container.auth_service()
        auth_service2 = container.auth_service()
        self.assertIsInstance(auth_service1, AuthService)
        self.assertIsInstance(auth_service2, AuthService)
        self.assertIsNot(auth_service1, auth_service2)

        self.assertIsInstance(auth_service1.db, sqlite3.Connection)
        self.assertIsInstance(auth_service2.db, sqlite3.Connection)
        self.assertIs(auth_service1.db, auth_service2.db)
        self.assertIs(auth_service1.db, container.database_client())
        self.assertIs(auth_service2.db, container.database_client())

        self.assertEqual(auth_service1.token_ttl, 3600)
        self.assertEqual(auth_service2.token_ttl, 3600)

        # Photo service
        photo_service1 = container.photo_service()
        photo_service2 = container.photo_service()
        self.assertIsInstance(photo_service1, PhotoService)
        self.assertIsInstance(photo_service2, PhotoService)
        self.assertIsNot(photo_service1, photo_service2)

        self.assertIsInstance(photo_service1.db, sqlite3.Connection)
        self.assertIsInstance(photo_service2.db, sqlite3.Connection)
        self.assertIs(photo_service1.db, photo_service2.db)
        self.assertIs(photo_service1.db, container.database_client())
        self.assertIs(photo_service2.db, container.database_client())

        self.assertIs(photo_service1.s3, photo_service2.s3)
        self.assertIs(photo_service1.s3, container.s3_client())
        self.assertIs(photo_service2.s3, container.s3_client())


class TestSchemaMultipleContainers(unittest.TestCase):

    def test(self):
        container = containers.DynamicContainer()
        container.from_yaml_schema(f'{_SAMPLES_DIR}/schemasample/container-multiple.yml')
        container.core.config.from_dict({
            'database': {
                'dsn': ':memory:',
            },
            'aws': {
                'access_key_id': 'KEY',
                'secret_access_key': 'SECRET',
            },
            'auth': {
                'token_ttl': 3600,
            },
        })

        # User service
        user_service1 = container.services.user()
        user_service2 = container.services.user()
        self.assertIsInstance(user_service1, UserService)
        self.assertIsInstance(user_service2, UserService)
        self.assertIsNot(user_service1, user_service2)

        self.assertIsInstance(user_service1.db, sqlite3.Connection)
        self.assertIsInstance(user_service2.db, sqlite3.Connection)
        self.assertIs(user_service1.db, user_service2.db)

        # Auth service
        auth_service1 = container.services.auth()
        auth_service2 = container.services.auth()
        self.assertIsInstance(auth_service1, AuthService)
        self.assertIsInstance(auth_service2, AuthService)
        self.assertIsNot(auth_service1, auth_service2)

        self.assertIsInstance(auth_service1.db, sqlite3.Connection)
        self.assertIsInstance(auth_service2.db, sqlite3.Connection)
        self.assertIs(auth_service1.db, auth_service2.db)
        self.assertIs(auth_service1.db, container.gateways.database_client())
        self.assertIs(auth_service2.db, container.gateways.database_client())

        self.assertEqual(auth_service1.token_ttl, 3600)
        self.assertEqual(auth_service2.token_ttl, 3600)

        # Photo service
        photo_service1 = container.services.photo()
        photo_service2 = container.services.photo()
        self.assertIsInstance(photo_service1, PhotoService)
        self.assertIsInstance(photo_service2, PhotoService)
        self.assertIsNot(photo_service1, photo_service2)

        self.assertIsInstance(photo_service1.db, sqlite3.Connection)
        self.assertIsInstance(photo_service2.db, sqlite3.Connection)
        self.assertIs(photo_service1.db, photo_service2.db)
        self.assertIs(photo_service1.db, container.gateways.database_client())
        self.assertIs(photo_service2.db, container.gateways.database_client())

        self.assertIs(photo_service1.s3, photo_service2.s3)
        self.assertIs(photo_service1.s3, container.gateways.s3_client())
        self.assertIs(photo_service2.s3, container.gateways.s3_client())


class TestSchemaMultipleContainersReordered(unittest.TestCase):

    def test(self):
        container = containers.DynamicContainer()
        container.from_yaml_schema(f'{_SAMPLES_DIR}/schemasample/container-multiple-reordered.yml')
        container.core.config.from_dict({
            'database': {
                'dsn': ':memory:',
            },
            'aws': {
                'access_key_id': 'KEY',
                'secret_access_key': 'SECRET',
            },
            'auth': {
                'token_ttl': 3600,
            },
        })

        # User service
        user_service1 = container.services.user()
        user_service2 = container.services.user()
        self.assertIsInstance(user_service1, UserService)
        self.assertIsInstance(user_service2, UserService)
        self.assertIsNot(user_service1, user_service2)

        self.assertIsInstance(user_service1.db, sqlite3.Connection)
        self.assertIsInstance(user_service2.db, sqlite3.Connection)
        self.assertIs(user_service1.db, user_service2.db)

        # Auth service
        auth_service1 = container.services.auth()
        auth_service2 = container.services.auth()
        self.assertIsInstance(auth_service1, AuthService)
        self.assertIsInstance(auth_service2, AuthService)
        self.assertIsNot(auth_service1, auth_service2)

        self.assertIsInstance(auth_service1.db, sqlite3.Connection)
        self.assertIsInstance(auth_service2.db, sqlite3.Connection)
        self.assertIs(auth_service1.db, auth_service2.db)
        self.assertIs(auth_service1.db, container.gateways.database_client())
        self.assertIs(auth_service2.db, container.gateways.database_client())

        self.assertEqual(auth_service1.token_ttl, 3600)
        self.assertEqual(auth_service2.token_ttl, 3600)

        # Photo service
        photo_service1 = container.services.photo()
        photo_service2 = container.services.photo()
        self.assertIsInstance(photo_service1, PhotoService)
        self.assertIsInstance(photo_service2, PhotoService)
        self.assertIsNot(photo_service1, photo_service2)

        self.assertIsInstance(photo_service1.db, sqlite3.Connection)
        self.assertIsInstance(photo_service2.db, sqlite3.Connection)
        self.assertIs(photo_service1.db, photo_service2.db)
        self.assertIs(photo_service1.db, container.gateways.database_client())
        self.assertIs(photo_service2.db, container.gateways.database_client())

        self.assertIs(photo_service1.s3, photo_service2.s3)
        self.assertIs(photo_service1.s3, container.gateways.s3_client())
        self.assertIs(photo_service2.s3, container.gateways.s3_client())


class TestSchemaMultipleContainersWithInlineProviders(unittest.TestCase):

    def test(self):
        container = containers.DynamicContainer()
        container.from_yaml_schema(f'{_SAMPLES_DIR}/schemasample/container-multiple-inline.yml')
        container.core.config.from_dict({
            'database': {
                'dsn': ':memory:',
            },
            'aws': {
                'access_key_id': 'KEY',
                'secret_access_key': 'SECRET',
            },
            'auth': {
                'token_ttl': 3600,
            },
        })

        # User service
        user_service1 = container.services.user()
        user_service2 = container.services.user()
        self.assertIsInstance(user_service1, UserService)
        self.assertIsInstance(user_service2, UserService)
        self.assertIsNot(user_service1, user_service2)

        self.assertIsInstance(user_service1.db, sqlite3.Connection)
        self.assertIsInstance(user_service2.db, sqlite3.Connection)
        self.assertIs(user_service1.db, user_service2.db)

        # Auth service
        auth_service1 = container.services.auth()
        auth_service2 = container.services.auth()
        self.assertIsInstance(auth_service1, AuthService)
        self.assertIsInstance(auth_service2, AuthService)
        self.assertIsNot(auth_service1, auth_service2)

        self.assertIsInstance(auth_service1.db, sqlite3.Connection)
        self.assertIsInstance(auth_service2.db, sqlite3.Connection)
        self.assertIs(auth_service1.db, auth_service2.db)
        self.assertIs(auth_service1.db, container.gateways.database_client())
        self.assertIs(auth_service2.db, container.gateways.database_client())

        self.assertEqual(auth_service1.token_ttl, 3600)
        self.assertEqual(auth_service2.token_ttl, 3600)

        # Photo service
        photo_service1 = container.services.photo()
        photo_service2 = container.services.photo()
        self.assertIsInstance(photo_service1, PhotoService)
        self.assertIsInstance(photo_service2, PhotoService)
        self.assertIsNot(photo_service1, photo_service2)

        self.assertIsInstance(photo_service1.db, sqlite3.Connection)
        self.assertIsInstance(photo_service2.db, sqlite3.Connection)
        self.assertIs(photo_service1.db, photo_service2.db)
        self.assertIs(photo_service1.db, container.gateways.database_client())
        self.assertIs(photo_service2.db, container.gateways.database_client())

        self.assertIs(photo_service1.s3, photo_service2.s3)
        self.assertIs(photo_service1.s3, container.gateways.s3_client())
        self.assertIs(photo_service2.s3, container.gateways.s3_client())


class TestSchemaBoto3Session(unittest.TestCase):

    @unittest.skip('Boto3 tries to connect to the internet')
    def test(self):
        container = containers.DynamicContainer()
        container.from_yaml_schema(f'{_SAMPLES_DIR}/schemasample/container-boto3-session.yml')
        container.config.from_dict(
            {
                'aws_access_key_id': 'key',
                'aws_secret_access_key': 'secret',
                'aws_session_token': 'token',
                'aws_region_name': 'us-east-1',
            },
        )

        self.assertEqual(container.s3_client().__class__.__name__, 'S3')
        self.assertEqual(container.sqs_client().__class__.__name__, 'SQS')



"""Dependency injector callable providers unit tests."""

import sys

import unittest

from dependency_injector import (
    providers,
    errors,
)


def _example(arg1, arg2, arg3, arg4):
    return arg1, arg2, arg3, arg4


class CallableTests(unittest.TestCase):

    def test_init_with_callable(self):
        self.assertTrue(providers.Callable(_example))

    def test_init_with_not_callable(self):
        self.assertRaises(errors.Error, providers.Callable, 123)

    def test_init_optional_provides(self):
        provider = providers.Callable()
        provider.set_provides(object)
        self.assertIs(provider.provides, object)
        self.assertIsInstance(provider(), object)

    def test_set_provides_returns_self(self):
        provider = providers.Callable()
        self.assertIs(provider.set_provides(object), provider)

    def test_provided_instance_provider(self):
        provider = providers.Callable(_example)
        self.assertIsInstance(provider.provided, providers.ProvidedInstance)

    def test_call(self):
        provider = providers.Callable(lambda: True)
        self.assertTrue(provider())

    def test_call_with_positional_args(self):
        provider = providers.Callable(_example,
                                      1, 2, 3, 4)
        self.assertTupleEqual(provider(), (1, 2, 3, 4))

    def test_call_with_keyword_args(self):
        provider = providers.Callable(_example,
                                      arg1=1, arg2=2, arg3=3, arg4=4)
        self.assertTupleEqual(provider(), (1, 2, 3, 4))

    def test_call_with_positional_and_keyword_args(self):
        provider = providers.Callable(_example,
                                      1, 2,
                                      arg3=3, arg4=4)
        self.assertTupleEqual(provider(), (1, 2, 3, 4))

    def test_call_with_context_args(self):
        provider = providers.Callable(_example, 1, 2)
        self.assertTupleEqual(provider(3, 4), (1, 2, 3, 4))

    def test_call_with_context_kwargs(self):
        provider = providers.Callable(_example, arg1=1)
        self.assertTupleEqual(provider(arg2=2, arg3=3, arg4=4), (1, 2, 3, 4))

    def test_call_with_context_args_and_kwargs(self):
        provider = providers.Callable(_example, 1)
        self.assertTupleEqual(provider(2, arg3=3, arg4=4), (1, 2, 3, 4))

    def test_fluent_interface(self):
        provider = providers.Singleton(_example) \
            .add_args(1, 2) \
            .add_kwargs(arg3=3, arg4=4)

        self.assertTupleEqual(provider(), (1, 2, 3, 4))

    def test_set_args(self):
        provider = providers.Callable(_example) \
            .add_args(1, 2) \
            .set_args(3, 4)
        self.assertEqual(provider.args, (3, 4))

    def test_set_kwargs(self):
        provider = providers.Callable(_example) \
            .add_kwargs(init_arg3=3, init_arg4=4) \
            .set_kwargs(init_arg3=4, init_arg4=5)
        self.assertEqual(provider.kwargs, dict(init_arg3=4, init_arg4=5))

    def test_clear_args(self):
        provider = providers.Callable(_example) \
            .add_args(1, 2) \
            .clear_args()
        self.assertEqual(provider.args, tuple())

    def test_clear_kwargs(self):
        provider = providers.Callable(_example) \
            .add_kwargs(init_arg3=3, init_arg4=4) \
            .clear_kwargs()
        self.assertEqual(provider.kwargs, dict())

    def test_call_overridden(self):
        provider = providers.Callable(_example)

        provider.override(providers.Object((4, 3, 2, 1)))
        provider.override(providers.Object((1, 2, 3, 4)))

        self.assertTupleEqual(provider(), (1, 2, 3, 4))

    def test_deepcopy(self):
        provider = providers.Callable(_example)

        provider_copy = providers.deepcopy(provider)

        self.assertIsNot(provider, provider_copy)
        self.assertIs(provider.provides, provider_copy.provides)
        self.assertIsInstance(provider, providers.Callable)

    def test_deepcopy_from_memo(self):
        provider = providers.Callable(_example)
        provider_copy_memo = providers.Callable(_example)

        provider_copy = providers.deepcopy(
            provider, memo={id(provider): provider_copy_memo})

        self.assertIs(provider_copy, provider_copy_memo)

    def test_deepcopy_args(self):
        provider = providers.Callable(_example)
        dependent_provider1 = providers.Callable(list)
        dependent_provider2 = providers.Callable(dict)

        provider.add_args(dependent_provider1, dependent_provider2)

        provider_copy = providers.deepcopy(provider)
        dependent_provider_copy1 = provider_copy.args[0]
        dependent_provider_copy2 = provider_copy.args[1]

        self.assertNotEqual(provider.args, provider_copy.args)

        self.assertIs(dependent_provider1.provides,
                      dependent_provider_copy1.provides)
        self.assertIsNot(dependent_provider1, dependent_provider_copy1)

        self.assertIs(dependent_provider2.provides,
                      dependent_provider_copy2.provides)
        self.assertIsNot(dependent_provider2, dependent_provider_copy2)

    def test_deepcopy_kwargs(self):
        provider = providers.Callable(_example)
        dependent_provider1 = providers.Callable(list)
        dependent_provider2 = providers.Callable(dict)

        provider.add_kwargs(a1=dependent_provider1, a2=dependent_provider2)

        provider_copy = providers.deepcopy(provider)
        dependent_provider_copy1 = provider_copy.kwargs['a1']
        dependent_provider_copy2 = provider_copy.kwargs['a2']

        self.assertNotEqual(provider.kwargs, provider_copy.kwargs)

        self.assertIs(dependent_provider1.provides,
                      dependent_provider_copy1.provides)
        self.assertIsNot(dependent_provider1, dependent_provider_copy1)

        self.assertIs(dependent_provider2.provides,
                      dependent_provider_copy2.provides)
        self.assertIsNot(dependent_provider2, dependent_provider_copy2)

    def test_deepcopy_overridden(self):
        provider = providers.Callable(_example)
        object_provider = providers.Object(object())

        provider.override(object_provider)

        provider_copy = providers.deepcopy(provider)
        object_provider_copy = provider_copy.overridden[0]

        self.assertIsNot(provider, provider_copy)
        self.assertIs(provider.provides, provider_copy.provides)
        self.assertIsInstance(provider, providers.Callable)

        self.assertIsNot(object_provider, object_provider_copy)
        self.assertIsInstance(object_provider_copy, providers.Object)

    def test_deepcopy_with_sys_streams(self):
        provider = providers.Callable(_example)
        provider.add_args(sys.stdin)
        provider.add_kwargs(a2=sys.stdout)

        provider_copy = providers.deepcopy(provider)

        self.assertIsNot(provider, provider_copy)
        self.assertIsInstance(provider_copy, providers.Callable)
        self.assertIs(provider.args[0], sys.stdin)
        self.assertIs(provider.kwargs['a2'], sys.stdout)

    def test_repr(self):
        provider = providers.Callable(_example)

        self.assertEqual(repr(provider),
                         '<dependency_injector.providers.'
                         'Callable({0}) at {1}>'.format(
                             repr(_example),
                             hex(id(provider))))


class DelegatedCallableTests(unittest.TestCase):

    def test_inheritance(self):
        self.assertIsInstance(providers.DelegatedCallable(_example),
                              providers.Callable)

    def test_is_provider(self):
        self.assertTrue(
            providers.is_provider(providers.DelegatedCallable(_example)))

    def test_is_delegated_provider(self):
        provider = providers.DelegatedCallable(_example)
        self.assertTrue(providers.is_delegated(provider))

    def test_repr(self):
        provider = providers.DelegatedCallable(_example)

        self.assertEqual(repr(provider),
                         '<dependency_injector.providers.'
                         'DelegatedCallable({0}) at {1}>'.format(
                             repr(_example),
                             hex(id(provider))))


class AbstractCallableTests(unittest.TestCase):

    def test_inheritance(self):
        self.assertIsInstance(providers.AbstractCallable(_example),
                              providers.Callable)

    def test_call_overridden_by_callable(self):
        def _abstract_example():
            pass

        provider = providers.AbstractCallable(_abstract_example)
        provider.override(providers.Callable(_example))

        self.assertTrue(provider(1, 2, 3, 4), (1, 2, 3, 4))

    def test_call_overridden_by_delegated_callable(self):
        def _abstract_example():
            pass

        provider = providers.AbstractCallable(_abstract_example)
        provider.override(providers.DelegatedCallable(_example))

        self.assertTrue(provider(1, 2, 3, 4), (1, 2, 3, 4))

    def test_call_not_overridden(self):
        provider = providers.AbstractCallable(_example)

        with self.assertRaises(errors.Error):
            provider(1, 2, 3, 4)

    def test_override_by_not_callable(self):
        provider = providers.AbstractCallable(_example)

        with self.assertRaises(errors.Error):
            provider.override(providers.Factory(object))

    def test_provide_not_implemented(self):
        provider = providers.AbstractCallable(_example)

        with self.assertRaises(NotImplementedError):
            provider._provide((1, 2, 3, 4), dict())

    def test_repr(self):
        provider = providers.AbstractCallable(_example)

        self.assertEqual(repr(provider),
                         '<dependency_injector.providers.'
                         'AbstractCallable({0}) at {1}>'.format(
                             repr(_example),
                             hex(id(provider))))


class CallableDelegateTests(unittest.TestCase):

    def setUp(self):
        self.delegated = providers.Callable(_example)
        self.delegate = providers.CallableDelegate(self.delegated)

    def test_is_delegate(self):
        self.assertIsInstance(self.delegate, providers.Delegate)

    def test_init_with_not_callable(self):
        self.assertRaises(errors.Error,
                          providers.CallableDelegate,
                          providers.Object(object()))



"""Dependency injector container provider unit tests."""

import copy

import unittest

from dependency_injector import containers, providers, errors


TEST_VALUE_1 = 'core_section_value1'
TEST_CONFIG_1 = {
    'core': {
        'section': {
            'value': TEST_VALUE_1,
        },
    },
}

TEST_VALUE_2 = 'core_section_value2'
TEST_CONFIG_2 = {
    'core': {
        'section': {
            'value': TEST_VALUE_2,
        },
    },
}


def _copied(value):
    return copy.deepcopy(value)


class TestCore(containers.DeclarativeContainer):
    config = providers.Configuration('core')
    value_getter = providers.Callable(lambda _: _, config.section.value)


class TestApplication(containers.DeclarativeContainer):
    config = providers.Configuration('config')
    core = providers.Container(TestCore, config=config.core)
    dict_factory = providers.Factory(dict, value=core.value_getter)


class ContainerTests(unittest.TestCase):

    def test(self):
        application = TestApplication(config=_copied(TEST_CONFIG_1))
        self.assertEqual(application.dict_factory(), {'value': TEST_VALUE_1})

    def test_double_override(self):
        application = TestApplication()
        application.config.override(_copied(TEST_CONFIG_1))
        application.config.override(_copied(TEST_CONFIG_2))
        self.assertEqual(application.dict_factory(), {'value': TEST_VALUE_2})

    def test_override(self):
        # See: https://github.com/ets-labs/python-dependency-injector/issues/354
        class D(containers.DeclarativeContainer):
            foo = providers.Object('foo')

        class A(containers.DeclarativeContainer):
            d = providers.DependenciesContainer()
            bar = providers.Callable(lambda f: f + '++', d.foo.provided)

        class B(containers.DeclarativeContainer):
            d = providers.Container(D)

            a = providers.Container(A, d=d)

        b = B(d=D())
        result = b.a().bar()
        self.assertEqual(result, 'foo++')

    def test_override_not_root_provider(self):
        # See: https://github.com/ets-labs/python-dependency-injector/issues/379
        class NestedContainer(containers.DeclarativeContainer):
            settings = providers.Configuration()

            print_settings = providers.Callable(
                lambda s: s,
                settings,
            )

        class TestContainer(containers.DeclarativeContainer):
            settings = providers.Configuration()

            root_container = providers.Container(
                NestedContainer,
                settings=settings,
            )

            not_root_container = providers.Selector(
                settings.container,
                using_factory=providers.Factory(
                    NestedContainer,
                    settings=settings,
                ),
                using_container=providers.Container(
                    NestedContainer,
                    settings=settings,
                )
            )

        container_using_factory = TestContainer(settings=dict(
            container='using_factory',
            foo='bar'
        ))
        self.assertEqual(
            container_using_factory.root_container().print_settings(),
            {'container': 'using_factory', 'foo': 'bar'},
        )
        self.assertEqual(
            container_using_factory.not_root_container().print_settings(),
            {'container': 'using_factory', 'foo': 'bar'},
        )


        container_using_container = TestContainer(settings=dict(
            container='using_container',
            foo='bar'
        ))
        self.assertEqual(
            container_using_container.root_container().print_settings(),
            {'container': 'using_container', 'foo': 'bar'},
        )
        self.assertEqual(
            container_using_container.not_root_container().print_settings(),
            {'container': 'using_container', 'foo': 'bar'},
        )

    def test_override_by_not_a_container(self):
        provider = providers.Container(TestCore)

        with self.assertRaises(errors.Error):
            provider.override(providers.Object('foo'))

    def test_lazy_overriding(self):
        # See: https://github.com/ets-labs/python-dependency-injector/issues/354

        class D(containers.DeclarativeContainer):
            foo = providers.Object("foo")

        class A(containers.DeclarativeContainer):
            d = providers.DependenciesContainer()
            bar = providers.Callable(lambda f: f + "++", d.foo.provided)

        class B(containers.DeclarativeContainer):
            d = providers.DependenciesContainer()

            a = providers.Container(A, d=d)

        b = B(d=D())
        result = b.a().bar()
        self.assertEqual(result, 'foo++')

    def test_lazy_overriding_deep(self):
        # Extended version of test_lazy_overriding()

        class D(containers.DeclarativeContainer):
            foo = providers.Object("foo")

        class C(containers.DeclarativeContainer):
            d = providers.DependenciesContainer()
            bar = providers.Callable(lambda f: f + "++", d.foo.provided)

        class A(containers.DeclarativeContainer):
            d = providers.DependenciesContainer()
            c = providers.Container(C, d=d)

        class B(containers.DeclarativeContainer):
            d = providers.DependenciesContainer()

            a = providers.Container(A, d=d)

        b = B(d=D())
        result = b.a().c().bar()
        self.assertEqual(result, 'foo++')

    def test_reset_last_overriding(self):
        application = TestApplication(config=_copied(TEST_CONFIG_1))
        application.core.override(TestCore(config=_copied(TEST_CONFIG_2['core'])))

        application.core.reset_last_overriding()

        self.assertEqual(application.dict_factory(), {'value': TEST_VALUE_1})

    def test_reset_last_overriding_only_overridden(self):
        application = TestApplication(config=_copied(TEST_CONFIG_1))
        application.core.override(providers.DependenciesContainer(config=_copied(TEST_CONFIG_2['core'])))

        application.core.reset_last_overriding()

        self.assertEqual(application.dict_factory(), {'value': TEST_VALUE_1})

    def test_override_context_manager(self):
        application = TestApplication(config=_copied(TEST_CONFIG_1))
        overriding_core = TestCore(config=_copied(TEST_CONFIG_2['core']))

        with application.core.override(overriding_core) as context_core:
            self.assertEqual(application.dict_factory(), {'value': TEST_VALUE_2})
            self.assertIs(context_core(), overriding_core)

        self.assertEqual(application.dict_factory(), {'value': TEST_VALUE_1})

    def test_reset_override(self):
        application = TestApplication(config=_copied(TEST_CONFIG_1))
        application.core.override(TestCore(config=_copied(TEST_CONFIG_2['core'])))

        application.core.reset_override()

        self.assertEqual(application.dict_factory(), {'value': None})

    def test_reset_override_only_overridden(self):
        application = TestApplication(config=_copied(TEST_CONFIG_1))
        application.core.override(providers.DependenciesContainer(config=_copied(TEST_CONFIG_2['core'])))

        application.core.reset_override()

        self.assertEqual(application.dict_factory(), {'value': None})

    def test_assign_parent(self):
        parent = providers.DependenciesContainer()
        provider = providers.Container(TestCore)

        provider.assign_parent(parent)

        self.assertIs(provider.parent, parent)

    def test_parent_name(self):
        container = containers.DynamicContainer()
        provider = providers.Container(TestCore)
        container.name = provider
        self.assertEqual(provider.parent_name, 'name')

    def test_parent_name_with_deep_parenting(self):
        provider = providers.Container(TestCore)
        container = providers.DependenciesContainer(name=provider)
        _ = providers.DependenciesContainer(container=container)
        self.assertEqual(provider.parent_name, 'container.name')

    def test_parent_name_is_none(self):
        provider = providers.Container(TestCore)
        self.assertIsNone(provider.parent_name)

    def test_parent_deepcopy(self):
        container = containers.DynamicContainer()
        provider = providers.Container(TestCore)
        container.name = provider

        copied = providers.deepcopy(container)

        self.assertIs(container.name.parent, container)
        self.assertIs(copied.name.parent, copied)

        self.assertIsNot(container, copied)
        self.assertIsNot(container.name, copied.name)
        self.assertIsNot(container.name.parent, copied.name.parent)

    def test_resolve_provider_name(self):
        container = providers.Container(TestCore)
        self.assertEqual(container.resolve_provider_name(container.value_getter), 'value_getter')

    def test_resolve_provider_name_no_provider(self):
        container = providers.Container(TestCore)
        with self.assertRaises(errors.Error):
            container.resolve_provider_name(providers.Provider())



import asyncio
import random
import unittest

from dependency_injector import containers, providers, errors

# Runtime import to get asyncutils module
import os
_TOP_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        '../',
    )),
)
import sys
sys.path.append(_TOP_DIR)

from asyncutils import AsyncTestCase


RESOURCE1 = object()
RESOURCE2 = object()


async def init_resource(resource):
    await asyncio.sleep(random.randint(1, 10) / 1000)
    yield resource
    await asyncio.sleep(random.randint(1, 10) / 1000)


class Client:
    def __init__(self, resource1: object, resource2: object) -> None:
        self.resource1 = resource1
        self.resource2 = resource2


class Service:
    def __init__(self, client: Client) -> None:
        self.client = client


class Container(containers.DeclarativeContainer):
    resource1 = providers.Resource(init_resource, providers.Object(RESOURCE1))
    resource2 = providers.Resource(init_resource, providers.Object(RESOURCE2))

    client = providers.Factory(
        Client,
        resource1=resource1,
        resource2=resource2,
    )

    service = providers.Factory(
        Service,
        client=client,
    )


class FactoryTests(AsyncTestCase):

    def test_args_injection(self):
        class ContainerWithArgs(containers.DeclarativeContainer):
            resource1 = providers.Resource(init_resource, providers.Object(RESOURCE1))
            resource2 = providers.Resource(init_resource, providers.Object(RESOURCE2))

            client = providers.Factory(
                Client,
                resource1,
                resource2,
            )

            service = providers.Factory(
                Service,
                client,
            )

        container = ContainerWithArgs()

        client1 = self._run(container.client())
        client2 = self._run(container.client())

        self.assertIsInstance(client1, Client)
        self.assertIs(client1.resource1, RESOURCE1)
        self.assertIs(client1.resource2, RESOURCE2)

        self.assertIsInstance(client2, Client)
        self.assertIs(client2.resource1, RESOURCE1)
        self.assertIs(client2.resource2, RESOURCE2)

        service1 = self._run(container.service())
        service2 = self._run(container.service())

        self.assertIsInstance(service1, Service)
        self.assertIsInstance(service1.client, Client)
        self.assertIs(service1.client.resource1, RESOURCE1)
        self.assertIs(service1.client.resource2, RESOURCE2)

        self.assertIsInstance(service2, Service)
        self.assertIsInstance(service2.client, Client)
        self.assertIs(service2.client.resource1, RESOURCE1)
        self.assertIs(service2.client.resource2, RESOURCE2)

        self.assertIsNot(service1.client, service2.client)

    def test_kwargs_injection(self):
        container = Container()

        client1 = self._run(container.client())
        client2 = self._run(container.client())

        self.assertIsInstance(client1, Client)
        self.assertIs(client1.resource1, RESOURCE1)
        self.assertIs(client1.resource2, RESOURCE2)

        self.assertIsInstance(client2, Client)
        self.assertIs(client2.resource1, RESOURCE1)
        self.assertIs(client2.resource2, RESOURCE2)

        service1 = self._run(container.service())
        service2 = self._run(container.service())

        self.assertIsInstance(service1, Service)
        self.assertIsInstance(service1.client, Client)
        self.assertIs(service1.client.resource1, RESOURCE1)
        self.assertIs(service1.client.resource2, RESOURCE2)

        self.assertIsInstance(service2, Service)
        self.assertIsInstance(service2.client, Client)
        self.assertIs(service2.client.resource1, RESOURCE1)
        self.assertIs(service2.client.resource2, RESOURCE2)

        self.assertIsNot(service1.client, service2.client)

    def test_context_kwargs_injection(self):
        resource2_extra = object()

        container = Container()

        client1 = self._run(container.client(resource2=resource2_extra))
        client2 = self._run(container.client(resource2=resource2_extra))

        self.assertIsInstance(client1, Client)
        self.assertIs(client1.resource1, RESOURCE1)
        self.assertIs(client1.resource2, resource2_extra)

        self.assertIsInstance(client2, Client)
        self.assertIs(client2.resource1, RESOURCE1)
        self.assertIs(client2.resource2, resource2_extra)

    def test_args_kwargs_injection(self):
        class ContainerWithArgsAndKwArgs(containers.DeclarativeContainer):
            resource1 = providers.Resource(init_resource, providers.Object(RESOURCE1))
            resource2 = providers.Resource(init_resource, providers.Object(RESOURCE2))

            client = providers.Factory(
                Client,
                resource1,
                resource2=resource2,
            )

            service = providers.Factory(
                Service,
                client=client,
            )

        container = ContainerWithArgsAndKwArgs()

        client1 = self._run(container.client())
        client2 = self._run(container.client())

        self.assertIsInstance(client1, Client)
        self.assertIs(client1.resource1, RESOURCE1)
        self.assertIs(client1.resource2, RESOURCE2)

        self.assertIsInstance(client2, Client)
        self.assertIs(client2.resource1, RESOURCE1)
        self.assertIs(client2.resource2, RESOURCE2)

        service1 = self._run(container.service())
        service2 = self._run(container.service())

        self.assertIsInstance(service1, Service)
        self.assertIsInstance(service1.client, Client)
        self.assertIs(service1.client.resource1, RESOURCE1)
        self.assertIs(service1.client.resource2, RESOURCE2)

        self.assertIsInstance(service2, Service)
        self.assertIsInstance(service2.client, Client)
        self.assertIs(service2.client.resource1, RESOURCE1)
        self.assertIs(service2.client.resource2, RESOURCE2)

        self.assertIsNot(service1.client, service2.client)

    def test_injection_error(self):
        async def init_resource():
            raise Exception('Something went wrong')

        class Container(containers.DeclarativeContainer):
            resource_with_error = providers.Resource(init_resource)

            client = providers.Factory(
                Client,
                resource1=resource_with_error,
                resource2=None,
            )

        container = Container()

        with self.assertRaises(Exception) as context:
            self._run(container.client())
        self.assertEqual(str(context.exception), 'Something went wrong')

    def test_injection_runtime_error_async_provides(self):
        async def create_client(*args,  **kwargs):
            raise Exception('Something went wrong')

        class Container(containers.DeclarativeContainer):
            resource = providers.Resource(init_resource, providers.Object(RESOURCE1))

            client = providers.Factory(
                create_client,
                resource1=resource,
                resource2=None,
            )

        container = Container()

        with self.assertRaises(Exception) as context:
            self._run(container.client())
        self.assertEqual(str(context.exception), 'Something went wrong')

    def test_injection_call_error_async_provides(self):
        async def create_client():  # <-- no args defined
            ...

        class Container(containers.DeclarativeContainer):
            resource = providers.Resource(init_resource, providers.Object(RESOURCE1))

            client = providers.Factory(
                create_client,
                resource1=resource,
                resource2=None,
            )

        container = Container()

        with self.assertRaises(TypeError) as context:
            self._run(container.client())
        self.assertIn("create_client() got", str(context.exception))
        self.assertIn("unexpected keyword argument", str(context.exception))

    def test_attributes_injection(self):
        class ContainerWithAttributes(containers.DeclarativeContainer):
            resource1 = providers.Resource(init_resource, providers.Object(RESOURCE1))
            resource2 = providers.Resource(init_resource, providers.Object(RESOURCE2))

            client = providers.Factory(
                Client,
                resource1,
                resource2=None,
            )
            client.add_attributes(resource2=resource2)

            service = providers.Factory(
                Service,
                client=None,
            )
            service.add_attributes(client=client)

        container = ContainerWithAttributes()

        client1 = self._run(container.client())
        client2 = self._run(container.client())

        self.assertIsInstance(client1, Client)
        self.assertIs(client1.resource1, RESOURCE1)
        self.assertIs(client1.resource2, RESOURCE2)

        self.assertIsInstance(client2, Client)
        self.assertIs(client2.resource1, RESOURCE1)
        self.assertIs(client2.resource2, RESOURCE2)

        service1 = self._run(container.service())
        service2 = self._run(container.service())

        self.assertIsInstance(service1, Service)
        self.assertIsInstance(service1.client, Client)
        self.assertIs(service1.client.resource1, RESOURCE1)
        self.assertIs(service1.client.resource2, RESOURCE2)

        self.assertIsInstance(service2, Service)
        self.assertIsInstance(service2.client, Client)
        self.assertIs(service2.client.resource1, RESOURCE1)
        self.assertIs(service2.client.resource2, RESOURCE2)

        self.assertIsNot(service1.client, service2.client)

    def test_attributes_injection_attribute_error(self):
        class ClientWithException(Client):
            @property
            def attribute_set_error(self):
                return None

            @attribute_set_error.setter
            def attribute_set_error(self, value):
                raise Exception('Something went wrong')

        class Container(containers.DeclarativeContainer):
            resource = providers.Resource(init_resource, providers.Object(RESOURCE1))

            client = providers.Factory(
                ClientWithException,
                resource1=resource,
                resource2=resource,
            )
            client.add_attributes(attribute_set_error=123)

        container = Container()

        with self.assertRaises(Exception) as context:
            self._run(container.client())
        self.assertEqual(str(context.exception), 'Something went wrong')

    def test_attributes_injection_runtime_error(self):
        async def init_resource():
            raise Exception('Something went wrong')

        class Container(containers.DeclarativeContainer):
            resource = providers.Resource(init_resource)

            client = providers.Factory(
                Client,
                resource1=None,
                resource2=None,
            )
            client.add_attributes(resource1=resource)
            client.add_attributes(resource2=resource)

        container = Container()

        with self.assertRaises(Exception) as context:
            self._run(container.client())
        self.assertEqual(str(context.exception), 'Something went wrong')

    def test_async_instance_and_sync_attributes_injection(self):
        class ContainerWithAttributes(containers.DeclarativeContainer):
            resource1 = providers.Resource(init_resource, providers.Object(RESOURCE1))

            client = providers.Factory(
                Client,
                resource1,
                resource2=None,
            )
            client.add_attributes(resource2=providers.Object(RESOURCE2))

            service = providers.Factory(
                Service,
                client=None,
            )
            service.add_attributes(client=client)

        container = ContainerWithAttributes()

        client1 = self._run(container.client())
        client2 = self._run(container.client())

        self.assertIsInstance(client1, Client)
        self.assertIs(client1.resource1, RESOURCE1)
        self.assertIs(client1.resource2, RESOURCE2)

        self.assertIsInstance(client2, Client)
        self.assertIs(client2.resource1, RESOURCE1)
        self.assertIs(client2.resource2, RESOURCE2)

        service1 = self._run(container.service())
        service2 = self._run(container.service())

        self.assertIsInstance(service1, Service)
        self.assertIsInstance(service1.client, Client)
        self.assertIs(service1.client.resource1, RESOURCE1)
        self.assertIs(service1.client.resource2, RESOURCE2)

        self.assertIsInstance(service2, Service)
        self.assertIsInstance(service2.client, Client)
        self.assertIs(service2.client.resource1, RESOURCE1)
        self.assertIs(service2.client.resource2, RESOURCE2)

        self.assertIsNot(service1.client, service2.client)


class FactoryAggregateTests(AsyncTestCase):

    def test_async_mode(self):
        object1 = object()
        object2 = object()

        async def _get_object1():
            return object1

        def _get_object2():
            return object2

        provider = providers.FactoryAggregate(
            object1=providers.Factory(_get_object1),
            object2=providers.Factory(_get_object2),
        )

        self.assertTrue(provider.is_async_mode_undefined())

        created_object1 = self._run(provider('object1'))
        self.assertIs(created_object1, object1)
        self.assertTrue(provider.is_async_mode_enabled())

        created_object2 = self._run(provider('object2'))
        self.assertIs(created_object2, object2)


class SingletonTests(AsyncTestCase):

    def test_injections(self):
        class ContainerWithSingletons(containers.DeclarativeContainer):
            resource1 = providers.Resource(init_resource, providers.Object(RESOURCE1))
            resource2 = providers.Resource(init_resource, providers.Object(RESOURCE2))

            client = providers.Singleton(
                Client,
                resource1=resource1,
                resource2=resource2,
            )

            service = providers.Singleton(
                Service,
                client=client,
            )

        container = ContainerWithSingletons()

        client1 = self._run(container.client())
        client2 = self._run(container.client())

        self.assertIsInstance(client1, Client)
        self.assertIs(client1.resource1, RESOURCE1)
        self.assertIs(client1.resource2, RESOURCE2)

        self.assertIsInstance(client2, Client)
        self.assertIs(client2.resource1, RESOURCE1)
        self.assertIs(client2.resource2, RESOURCE2)

        service1 = self._run(container.service())
        service2 = self._run(container.service())

        self.assertIsInstance(service1, Service)
        self.assertIsInstance(service1.client, Client)
        self.assertIs(service1.client.resource1, RESOURCE1)
        self.assertIs(service1.client.resource2, RESOURCE2)

        self.assertIsInstance(service2, Service)
        self.assertIsInstance(service2.client, Client)
        self.assertIs(service2.client.resource1, RESOURCE1)
        self.assertIs(service2.client.resource2, RESOURCE2)

        self.assertIs(service1, service2)
        self.assertIs(service1.client, service2.client)
        self.assertIs(service1.client, client1)

        self.assertIs(service2.client, client2)
        self.assertIs(client1, client2)

    def test_async_mode(self):
        instance = object()

        async def create_instance():
            return instance

        provider = providers.Singleton(create_instance)

        instance1 = self._run(provider())
        instance2 = self._run(provider())

        self.assertIs(instance1, instance2)
        self.assertIs(instance, instance)

    def test_async_init_with_error(self):
        # Disable default exception handling to prevent output
        asyncio.get_event_loop().set_exception_handler(lambda loop, context: ...)

        async def create_instance():
            create_instance.counter += 1
            raise RuntimeError()

        create_instance.counter = 0

        provider = providers.Singleton(create_instance)


        future = provider()
        self.assertTrue(provider.is_async_mode_enabled())

        with self.assertRaises(RuntimeError):
            self._run(future)

        self.assertEqual(create_instance.counter, 1)
        self.assertTrue(provider.is_async_mode_enabled())

        with self.assertRaises(RuntimeError):
            self._run(provider())

        self.assertEqual(create_instance.counter, 2)
        self.assertTrue(provider.is_async_mode_enabled())

        # Restore default exception handling
        asyncio.get_event_loop().set_exception_handler(None)


class DelegatedSingletonTests(AsyncTestCase):

    def test_async_mode(self):
        instance = object()

        async def create_instance():
            return instance

        provider = providers.DelegatedSingleton(create_instance)

        instance1 = self._run(provider())
        instance2 = self._run(provider())

        self.assertIs(instance1, instance2)
        self.assertIs(instance, instance)


class ThreadSafeSingletonTests(AsyncTestCase):

    def test_async_mode(self):
        instance = object()

        async def create_instance():
            return instance

        provider = providers.ThreadSafeSingleton(create_instance)

        instance1 = self._run(provider())
        instance2 = self._run(provider())

        self.assertIs(instance1, instance2)
        self.assertIs(instance, instance)


class DelegatedThreadSafeSingletonTests(AsyncTestCase):

    def test_async_mode(self):
        instance = object()

        async def create_instance():
            return instance

        provider = providers.DelegatedThreadSafeSingleton(create_instance)

        instance1 = self._run(provider())
        instance2 = self._run(provider())

        self.assertIs(instance1, instance2)
        self.assertIs(instance, instance)


class ThreadLocalSingletonTests(AsyncTestCase):

    def test_async_mode(self):
        instance = object()

        async def create_instance():
            return instance

        provider = providers.ThreadLocalSingleton(create_instance)

        instance1 = self._run(provider())
        instance2 = self._run(provider())

        self.assertIs(instance1, instance2)
        self.assertIs(instance, instance)


    def test_async_init_with_error(self):
        # Disable default exception handling to prevent output
        asyncio.get_event_loop().set_exception_handler(lambda loop, context: ...)

        async def create_instance():
            create_instance.counter += 1
            raise RuntimeError()
        create_instance.counter = 0

        provider = providers.ThreadLocalSingleton(create_instance)

        future = provider()
        self.assertTrue(provider.is_async_mode_enabled())

        with self.assertRaises(RuntimeError):
            self._run(future)

        self.assertEqual(create_instance.counter, 1)
        self.assertTrue(provider.is_async_mode_enabled())

        with self.assertRaises(RuntimeError):
            self._run(provider())

        self.assertEqual(create_instance.counter, 2)
        self.assertTrue(provider.is_async_mode_enabled())

        # Restore default exception handling
        asyncio.get_event_loop().set_exception_handler(None)


class DelegatedThreadLocalSingletonTests(AsyncTestCase):

    def test_async_mode(self):
        instance = object()

        async def create_instance():
            return instance

        provider = providers.DelegatedThreadLocalSingleton(create_instance)

        instance1 = self._run(provider())
        instance2 = self._run(provider())

        self.assertIs(instance1, instance2)
        self.assertIs(instance, instance)


class ProvidedInstanceTests(AsyncTestCase):

    def test_provided_attribute(self):
        class TestClient:
            def __init__(self, resource):
                self.resource = resource

        class TestService:
            def __init__(self, resource):
                self.resource = resource

        class TestContainer(containers.DeclarativeContainer):
            resource = providers.Resource(init_resource, providers.Object(RESOURCE1))
            client = providers.Factory(TestClient, resource=resource)
            service = providers.Factory(TestService, resource=client.provided.resource)

        container = TestContainer()

        instance1, instance2 = self._run(
            asyncio.gather(
                container.service(),
                container.service(),
            ),
        )

        self.assertIs(instance1.resource, RESOURCE1)
        self.assertIs(instance2.resource, RESOURCE1)
        self.assertIs(instance1.resource, instance2.resource)

    def test_provided_attribute_error(self):
        async def raise_exception():
            raise RuntimeError()

        class TestContainer(containers.DeclarativeContainer):
            client = providers.Factory(raise_exception)

        container = TestContainer()

        with self.assertRaises(RuntimeError):
            self._run(container.client.provided.attr())

    def test_provided_attribute_undefined_attribute(self):
        class TestClient:
            def __init__(self, resource):
                self.resource = resource

        class TestContainer(containers.DeclarativeContainer):
            resource = providers.Resource(init_resource, providers.Object(RESOURCE1))
            client = providers.Factory(TestClient, resource=resource)

        container = TestContainer()

        with self.assertRaises(AttributeError):
            self._run(container.client.provided.attr())

    def test_provided_item(self):
        class TestClient:
            def __init__(self, resource):
                self.resource = resource

            def __getitem__(self, item):
                return getattr(self, item)

        class TestService:
            def __init__(self, resource):
                self.resource = resource

        class TestContainer(containers.DeclarativeContainer):
            resource = providers.Resource(init_resource, providers.Object(RESOURCE1))
            client = providers.Factory(TestClient, resource=resource)
            service = providers.Factory(TestService, resource=client.provided['resource'])

        container = TestContainer()

        instance1, instance2 = self._run(
            asyncio.gather(
                container.service(),
                container.service(),
            ),
        )

        self.assertIs(instance1.resource, RESOURCE1)
        self.assertIs(instance2.resource, RESOURCE1)
        self.assertIs(instance1.resource, instance2.resource)

    def test_provided_item_error(self):
        async def raise_exception():
            raise RuntimeError()

        class TestContainer(containers.DeclarativeContainer):
            client = providers.Factory(raise_exception)

        container = TestContainer()

        with self.assertRaises(RuntimeError):
            self._run(container.client.provided['item']())

    def test_provided_item_undefined_item(self):
        class TestContainer(containers.DeclarativeContainer):
            resource = providers.Resource(init_resource, providers.Object(RESOURCE1))
            client = providers.Factory(dict, resource=resource)

        container = TestContainer()

        with self.assertRaises(KeyError):
            self._run(container.client.provided['item']())

    def test_provided_method_call(self):
        class TestClient:
            def __init__(self, resource):
                self.resource = resource

            def get_resource(self):
                return self.resource

        class TestService:
            def __init__(self, resource):
                self.resource = resource

        class TestContainer(containers.DeclarativeContainer):
            resource = providers.Resource(init_resource, providers.Object(RESOURCE1))
            client = providers.Factory(TestClient, resource=resource)
            service = providers.Factory(TestService, resource=client.provided.get_resource.call())

        container = TestContainer()

        instance1, instance2 = self._run(
            asyncio.gather(
                container.service(),
                container.service(),
            ),
        )

        self.assertIs(instance1.resource, RESOURCE1)
        self.assertIs(instance2.resource, RESOURCE1)
        self.assertIs(instance1.resource, instance2.resource)

    def test_provided_method_call_parent_error(self):
        async def raise_exception():
            raise RuntimeError()

        class TestContainer(containers.DeclarativeContainer):
            client = providers.Factory(raise_exception)

        container = TestContainer()

        with self.assertRaises(RuntimeError):
            self._run(container.client.provided.method.call()())

    def test_provided_method_call_error(self):
        class TestClient:
            def method(self):
                raise RuntimeError()

        class TestContainer(containers.DeclarativeContainer):
            client = providers.Factory(TestClient)

        container = TestContainer()

        with self.assertRaises(RuntimeError):
            self._run(container.client.provided.method.call()())


class DependencyTests(AsyncTestCase):

    def test_provide_error(self):
        async def get_async():
            raise Exception

        provider = providers.Dependency()
        provider.override(providers.Callable(get_async))

        with self.assertRaises(Exception):
            self._run(provider())

    def test_isinstance(self):
        dependency = 1.0

        async def get_async():
            return dependency

        provider = providers.Dependency(instance_of=float)
        provider.override(providers.Callable(get_async))

        self.assertTrue(provider.is_async_mode_undefined())

        dependency1 = self._run(provider())

        self.assertTrue(provider.is_async_mode_enabled())

        dependency2 = self._run(provider())

        self.assertEqual(dependency1, dependency)
        self.assertEqual(dependency2, dependency)

    def test_isinstance_invalid(self):
        async def get_async():
            return {}

        provider = providers.Dependency(instance_of=float)
        provider.override(providers.Callable(get_async))

        self.assertTrue(provider.is_async_mode_undefined())

        with self.assertRaises(errors.Error):
            self._run(provider())

        self.assertTrue(provider.is_async_mode_enabled())

    def test_async_mode(self):
        dependency = 123

        async def get_async():
            return dependency

        def get_sync():
            return dependency

        provider = providers.Dependency(instance_of=int)
        provider.override(providers.Factory(get_async))

        self.assertTrue(provider.is_async_mode_undefined())

        dependency1 = self._run(provider())

        self.assertTrue(provider.is_async_mode_enabled())

        dependency2 = self._run(provider())
        self.assertEqual(dependency1, dependency)
        self.assertEqual(dependency2, dependency)

        provider.override(providers.Factory(get_sync))

        dependency3 = self._run(provider())

        self.assertTrue(provider.is_async_mode_enabled())

        dependency4 = self._run(provider())
        self.assertEqual(dependency3, dependency)
        self.assertEqual(dependency4, dependency)


class ListTests(AsyncTestCase):

    def test_provide(self):
        # See issue: https://github.com/ets-labs/python-dependency-injector/issues/450
        async def create_resource(param: str):
            return param

        class Container(containers.DeclarativeContainer):

            resources = providers.List(
                providers.Resource(create_resource, 'foo'),
                providers.Resource(create_resource, 'bar')
            )

        container = Container()
        resources = self._run(container.resources())

        self.assertEqual(resources[0], 'foo')
        self.assertEqual(resources[1], 'bar')


class DictTests(AsyncTestCase):

    def test_provide(self):
        async def create_resource(param: str):
            return param

        class Container(containers.DeclarativeContainer):

            resources = providers.Dict(
                foo=providers.Resource(create_resource, 'foo'),
                bar=providers.Resource(create_resource, 'bar')
            )

        container = Container()
        resources = self._run(container.resources())

        self.assertEqual(resources['foo'], 'foo')
        self.assertEqual(resources['bar'], 'bar')


class OverrideTests(AsyncTestCase):

    def test_provider(self):
        dependency = object()

        async def _get_dependency_async():
            return dependency

        def _get_dependency_sync():
            return dependency

        provider = providers.Provider()

        provider.override(providers.Callable(_get_dependency_async))
        dependency1 = self._run(provider())

        provider.override(providers.Callable(_get_dependency_sync))
        dependency2 = self._run(provider())

        self.assertIs(dependency1, dependency)
        self.assertIs(dependency2, dependency)

    def test_callable(self):
        dependency = object()

        async def _get_dependency_async():
            return dependency

        def _get_dependency_sync():
            return dependency

        provider = providers.Callable(_get_dependency_async)
        dependency1 = self._run(provider())

        provider.override(providers.Callable(_get_dependency_sync))
        dependency2 = self._run(provider())

        self.assertIs(dependency1, dependency)
        self.assertIs(dependency2, dependency)

    def test_factory(self):
        dependency = object()

        async def _get_dependency_async():
            return dependency

        def _get_dependency_sync():
            return dependency

        provider = providers.Factory(_get_dependency_async)
        dependency1 = self._run(provider())

        provider.override(providers.Callable(_get_dependency_sync))
        dependency2 = self._run(provider())

        self.assertIs(dependency1, dependency)
        self.assertIs(dependency2, dependency)

    def test_async_mode_enabling(self):
        dependency = object()

        async def _get_dependency_async():
            return dependency

        provider = providers.Callable(_get_dependency_async)
        self.assertTrue(provider.is_async_mode_undefined())

        self._run(provider())

        self.assertTrue(provider.is_async_mode_enabled())

    def test_async_mode_disabling(self):
        dependency = object()

        def _get_dependency():
            return dependency

        provider = providers.Callable(_get_dependency)
        self.assertTrue(provider.is_async_mode_undefined())

        provider()

        self.assertTrue(provider.is_async_mode_disabled())

    def test_async_mode_enabling_on_overriding(self):
        dependency = object()

        async def _get_dependency_async():
            return dependency

        provider = providers.Provider()
        provider.override(providers.Callable(_get_dependency_async))
        self.assertTrue(provider.is_async_mode_undefined())

        self._run(provider())

        self.assertTrue(provider.is_async_mode_enabled())

    def test_async_mode_disabling_on_overriding(self):
        dependency = object()

        def _get_dependency():
            return dependency

        provider = providers.Provider()
        provider.override(providers.Callable(_get_dependency))
        self.assertTrue(provider.is_async_mode_undefined())

        provider()

        self.assertTrue(provider.is_async_mode_disabled())


class TestAsyncModeApi(unittest.TestCase):

    def setUp(self):
        self.provider = providers.Provider()

    def test_default_mode(self):
        self.assertFalse(self.provider.is_async_mode_enabled())
        self.assertFalse(self.provider.is_async_mode_disabled())
        self.assertTrue(self.provider.is_async_mode_undefined())

    def test_enable(self):
        self.provider.enable_async_mode()

        self.assertTrue(self.provider.is_async_mode_enabled())
        self.assertFalse(self.provider.is_async_mode_disabled())
        self.assertFalse(self.provider.is_async_mode_undefined())

    def test_disable(self):
        self.provider.disable_async_mode()

        self.assertFalse(self.provider.is_async_mode_enabled())
        self.assertTrue(self.provider.is_async_mode_disabled())
        self.assertFalse(self.provider.is_async_mode_undefined())

    def test_reset(self):
        self.provider.enable_async_mode()

        self.assertTrue(self.provider.is_async_mode_enabled())
        self.assertFalse(self.provider.is_async_mode_disabled())
        self.assertFalse(self.provider.is_async_mode_undefined())

        self.provider.reset_async_mode()

        self.assertFalse(self.provider.is_async_mode_enabled())
        self.assertFalse(self.provider.is_async_mode_disabled())
        self.assertTrue(self.provider.is_async_mode_undefined())


class AsyncTypingStubTests(AsyncTestCase):

    def test_async_(self):
        container = Container()

        client1 = self._run(container.client.async_())
        client2 = self._run(container.client.async_())

        self.assertIsInstance(client1, Client)
        self.assertIs(client1.resource1, RESOURCE1)
        self.assertIs(client1.resource2, RESOURCE2)

        self.assertIsInstance(client2, Client)
        self.assertIs(client2.resource1, RESOURCE1)
        self.assertIs(client2.resource2, RESOURCE2)

        service1 = self._run(container.service.async_())
        service2 = self._run(container.service.async_())

        self.assertIsInstance(service1, Service)
        self.assertIsInstance(service1.client, Client)
        self.assertIs(service1.client.resource1, RESOURCE1)
        self.assertIs(service1.client.resource2, RESOURCE2)

        self.assertIsInstance(service2, Service)
        self.assertIsInstance(service2.client, Client)
        self.assertIs(service2.client.resource1, RESOURCE1)
        self.assertIs(service2.client.resource2, RESOURCE2)

        self.assertIsNot(service1.client, service2.client)


class AsyncProvidersWithAsyncDependenciesTests(AsyncTestCase):

    def test_injections(self):
        # See: https://github.com/ets-labs/python-dependency-injector/issues/368
        async def async_db_provider():
            return {'db': 'ok'}

        async def async_service(db=None):
            return {'service': 'ok', 'db': db}

        class Container(containers.DeclarativeContainer):

            db = providers.Factory(async_db_provider)
            service = providers.Singleton(async_service, db=db)

        container = Container()
        service = self._run(container.service())

        self.assertEqual(service, {'service': 'ok', 'db': {'db': 'ok'}})


class AsyncProviderWithAwaitableObjectTests(AsyncTestCase):

    def test(self):
        class SomeResource:
            def __await__(self):
                raise RuntimeError('Should never happen')

        async def init_resource():
            pool = SomeResource()
            yield pool

        class Service:
            def __init__(self, resource) -> None:
                self.resource = resource

        class Container(containers.DeclarativeContainer):
            resource = providers.Resource(init_resource)
            service = providers.Singleton(Service, resource=resource)

        container = Container()

        self._run(container.init_resources())
        self.assertIsInstance(container.service(), asyncio.Future)
        self.assertIsInstance(container.resource(), asyncio.Future)

        resource = self._run(container.resource())
        service = self._run(container.service())

        self.assertIsInstance(resource, SomeResource)
        self.assertIsInstance(service.resource, SomeResource)
        self.assertIs(service.resource, resource)

    def test_without_init_resources(self):
        class SomeResource:
            def __await__(self):
                raise RuntimeError('Should never happen')

        async def init_resource():
            pool = SomeResource()
            yield pool

        class Service:
            def __init__(self, resource) -> None:
                self.resource = resource

        class Container(containers.DeclarativeContainer):
            resource = providers.Resource(init_resource)
            service = providers.Singleton(Service, resource=resource)

        container = Container()

        self.assertIsInstance(container.service(), asyncio.Future)
        self.assertIsInstance(container.resource(), asyncio.Future)

        resource = self._run(container.resource())
        service = self._run(container.service())

        self.assertIsInstance(resource, SomeResource)
        self.assertIsInstance(service.resource, SomeResource)
        self.assertIs(service.resource, resource)



"""Dependency injector dynamic container unit tests."""

import unittest

from dependency_injector import (
    containers,
    providers,
    errors,
)


class ContainerA(containers.DeclarativeContainer):
    p11 = providers.Provider()
    p12 = providers.Provider()


class DeclarativeContainerInstanceTests(unittest.TestCase):

    def test_providers_attribute(self):
        container_a1 = ContainerA()
        container_a2 = ContainerA()

        self.assertIsNot(container_a1.p11, container_a2.p11)
        self.assertIsNot(container_a1.p12, container_a2.p12)
        self.assertNotEqual(container_a1.providers, container_a2.providers)

    def test_dependencies_attribute(self):
        container = ContainerA()
        container.a1 = providers.Dependency()
        container.a2 = providers.DependenciesContainer()
        self.assertEqual(container.dependencies, {'a1': container.a1, 'a2': container.a2})

    def test_set_get_del_providers(self):
        p13 = providers.Provider()

        container_a1 = ContainerA()
        container_a2 = ContainerA()

        container_a1.p13 = p13
        container_a2.p13 = p13

        self.assertEqual(ContainerA.providers, dict(p11=ContainerA.p11,
                                                    p12=ContainerA.p12))
        self.assertEqual(ContainerA.cls_providers, dict(p11=ContainerA.p11,
                                                        p12=ContainerA.p12))

        self.assertEqual(container_a1.providers, dict(p11=container_a1.p11,
                                                      p12=container_a1.p12,
                                                      p13=p13))
        self.assertEqual(container_a2.providers, dict(p11=container_a2.p11,
                                                      p12=container_a2.p12,
                                                      p13=p13))

        del container_a1.p13
        self.assertEqual(container_a1.providers, dict(p11=container_a1.p11,
                                                      p12=container_a1.p12))

        del container_a2.p13
        self.assertEqual(container_a2.providers, dict(p11=container_a2.p11,
                                                      p12=container_a2.p12))

        del container_a1.p11
        del container_a1.p12
        self.assertEqual(container_a1.providers, dict())
        self.assertEqual(ContainerA.providers, dict(p11=ContainerA.p11,
                                                    p12=ContainerA.p12))

        del container_a2.p11
        del container_a2.p12
        self.assertEqual(container_a2.providers, dict())
        self.assertEqual(ContainerA.providers, dict(p11=ContainerA.p11,
                                                    p12=ContainerA.p12))

    def test_set_invalid_provider_type(self):
        container_a = ContainerA()
        container_a.provider_type = providers.Object

        with self.assertRaises(errors.Error):
            container_a.px = providers.Provider()

        self.assertIs(ContainerA.provider_type,
                      containers.DeclarativeContainer.provider_type)

    def test_set_providers(self):
        p13 = providers.Provider()
        p14 = providers.Provider()
        container_a = ContainerA()

        container_a.set_providers(p13=p13, p14=p14)

        self.assertIs(container_a.p13, p13)
        self.assertIs(container_a.p14, p14)

    def test_override(self):
        class _Container(containers.DeclarativeContainer):
            p11 = providers.Provider()

        class _OverridingContainer1(containers.DeclarativeContainer):
            p11 = providers.Provider()

        class _OverridingContainer2(containers.DeclarativeContainer):
            p11 = providers.Provider()
            p12 = providers.Provider()

        container = _Container()
        overriding_container1 = _OverridingContainer1()
        overriding_container2 = _OverridingContainer2()

        container.override(overriding_container1)
        container.override(overriding_container2)

        self.assertEqual(container.overridden,
                         (overriding_container1,
                          overriding_container2))
        self.assertEqual(container.p11.overridden,
                         (overriding_container1.p11,
                          overriding_container2.p11))

        self.assertEqual(_Container.overridden, tuple())
        self.assertEqual(_Container.p11.overridden, tuple())

    def test_override_with_itself(self):
        container = ContainerA()
        with self.assertRaises(errors.Error):
            container.override(container)

    def test_override_providers(self):
        p1 = providers.Provider()
        p2 = providers.Provider()
        container_a = ContainerA()

        container_a.override_providers(p11=p1, p12=p2)

        self.assertIs(container_a.p11.last_overriding, p1)
        self.assertIs(container_a.p12.last_overriding, p2)

    def test_override_providers_with_unknown_provider(self):
        container_a = ContainerA()

        with self.assertRaises(AttributeError):
            container_a.override_providers(unknown=providers.Provider())

    def test_reset_last_overriding(self):
        class _Container(containers.DeclarativeContainer):
            p11 = providers.Provider()

        class _OverridingContainer1(containers.DeclarativeContainer):
            p11 = providers.Provider()

        class _OverridingContainer2(containers.DeclarativeContainer):
            p11 = providers.Provider()
            p12 = providers.Provider()

        container = _Container()
        overriding_container1 = _OverridingContainer1()
        overriding_container2 = _OverridingContainer2()

        container.override(overriding_container1)
        container.override(overriding_container2)
        container.reset_last_overriding()

        self.assertEqual(container.overridden,
                         (overriding_container1,))
        self.assertEqual(container.p11.overridden,
                         (overriding_container1.p11,))

    def test_reset_last_overriding_when_not_overridden(self):
        container = ContainerA()

        with self.assertRaises(errors.Error):
            container.reset_last_overriding()

    def test_reset_override(self):
        class _Container(containers.DeclarativeContainer):
            p11 = providers.Provider()

        class _OverridingContainer1(containers.DeclarativeContainer):
            p11 = providers.Provider()

        class _OverridingContainer2(containers.DeclarativeContainer):
            p11 = providers.Provider()
            p12 = providers.Provider()

        container = _Container()
        overriding_container1 = _OverridingContainer1()
        overriding_container2 = _OverridingContainer2()

        container.override(overriding_container1)
        container.override(overriding_container2)
        container.reset_override()

        self.assertEqual(container.overridden, tuple())
        self.assertEqual(container.p11.overridden, tuple())

    def test_init_and_shutdown_resources_ordering(self):
        """Test init and shutdown resources.

        Methods .init_resources() and .shutdown_resources() should respect resources dependencies.
        Initialization should first initialize resources without dependencies and then provide
        these resources to other resources. Resources shutdown should follow the same rule: first
        shutdown resources without initialized dependencies and then continue correspondingly
        until all resources are shutdown.
        """
        initialized_resources = []
        shutdown_resources = []

        def _resource(name, **_):
            initialized_resources.append(name)
            yield name
            shutdown_resources.append(name)

        class Container(containers.DeclarativeContainer):
            resource1 = providers.Resource(
                _resource,
                name='r1',
            )
            resource2 = providers.Resource(
                _resource,
                name='r2',
                r1=resource1,
            )
            resource3 = providers.Resource(
                _resource,
                name='r3',
                r2=resource2,
            )

        container = Container()

        container.init_resources()
        self.assertEqual(initialized_resources, ['r1', 'r2', 'r3'])
        self.assertEqual(shutdown_resources, [])

        container.shutdown_resources()
        self.assertEqual(initialized_resources, ['r1', 'r2', 'r3'])
        self.assertEqual(shutdown_resources, ['r3', 'r2', 'r1'])

        container.init_resources()
        self.assertEqual(initialized_resources, ['r1', 'r2', 'r3', 'r1', 'r2', 'r3'])
        self.assertEqual(shutdown_resources, ['r3', 'r2', 'r1'])

        container.shutdown_resources()
        self.assertEqual(initialized_resources, ['r1', 'r2', 'r3', 'r1', 'r2', 'r3'])
        self.assertEqual(shutdown_resources, ['r3', 'r2', 'r1', 'r3', 'r2', 'r1'])

    def test_shutdown_resources_circular_dependencies_breaker(self):
        def _resource(name, **_):
            yield name

        class Container(containers.DeclarativeContainer):
            resource1 = providers.Resource(
                _resource,
                name='r1',
            )
            resource2 = providers.Resource(
                _resource,
                name='r2',
                r1=resource1,
            )
            resource3 = providers.Resource(
                _resource,
                name='r3',
                r2=resource2,
            )

        container = Container()
        container.init_resources()

        # Create circular dependency after initialization (r3 -> r2 -> r1 -> r3 -> ...)
        container.resource1.add_kwargs(r3=container.resource3)

        with self.assertRaises(RuntimeError) as context:
            container.shutdown_resources()
        self.assertEqual(str(context.exception), 'Unable to resolve resources shutdown order')

    def test_init_shutdown_nested_resources(self):
        def _init1():
            _init1.init_counter += 1
            yield
            _init1.shutdown_counter += 1

        _init1.init_counter = 0
        _init1.shutdown_counter = 0

        def _init2():
            _init2.init_counter += 1
            yield
            _init2.shutdown_counter += 1

        _init2.init_counter = 0
        _init2.shutdown_counter = 0

        class Container(containers.DeclarativeContainer):

            service = providers.Factory(
                dict,
                resource1=providers.Resource(_init1),
                resource2=providers.Resource(_init2),
            )

        container = Container()
        self.assertEqual(_init1.init_counter, 0)
        self.assertEqual(_init1.shutdown_counter, 0)
        self.assertEqual(_init2.init_counter, 0)
        self.assertEqual(_init2.shutdown_counter, 0)

        container.init_resources()
        self.assertEqual(_init1.init_counter, 1)
        self.assertEqual(_init1.shutdown_counter, 0)
        self.assertEqual(_init2.init_counter, 1)
        self.assertEqual(_init2.shutdown_counter, 0)

        container.shutdown_resources()
        self.assertEqual(_init1.init_counter, 1)
        self.assertEqual(_init1.shutdown_counter, 1)
        self.assertEqual(_init2.init_counter, 1)
        self.assertEqual(_init2.shutdown_counter, 1)

        container.init_resources()
        container.shutdown_resources()
        self.assertEqual(_init1.init_counter, 2)
        self.assertEqual(_init1.shutdown_counter, 2)
        self.assertEqual(_init2.init_counter, 2)
        self.assertEqual(_init2.shutdown_counter, 2)

    def test_reset_singletons(self):
        class SubSubContainer(containers.DeclarativeContainer):
            singleton = providers.Singleton(object)

        class SubContainer(containers.DeclarativeContainer):
            singleton = providers.Singleton(object)
            sub_sub_container = providers.Container(SubSubContainer)

        class Container(containers.DeclarativeContainer):
            singleton = providers.Singleton(object)
            sub_container = providers.Container(SubContainer)

        container = Container()

        obj11 = container.singleton()
        obj12 = container.sub_container().singleton()
        obj13 = container.sub_container().sub_sub_container().singleton()

        obj21 = container.singleton()
        obj22 = container.sub_container().singleton()
        obj23 = container.sub_container().sub_sub_container().singleton()

        self.assertIs(obj11, obj21)
        self.assertIs(obj12, obj22)
        self.assertIs(obj13, obj23)

        container.reset_singletons()

        obj31 = container.singleton()
        obj32 = container.sub_container().singleton()
        obj33 = container.sub_container().sub_sub_container().singleton()

        obj41 = container.singleton()
        obj42 = container.sub_container().singleton()
        obj43 = container.sub_container().sub_sub_container().singleton()

        self.assertIsNot(obj11, obj31)
        self.assertIsNot(obj12, obj32)
        self.assertIsNot(obj13, obj33)

        self.assertIsNot(obj21, obj31)
        self.assertIsNot(obj22, obj32)
        self.assertIsNot(obj23, obj33)

        self.assertIs(obj31, obj41)
        self.assertIs(obj32, obj42)
        self.assertIs(obj33, obj43)

    def test_reset_singletons_context_manager(self):
        class Item:
            def __init__(self, dependency):
                self.dependency = dependency

        class Container(containers.DeclarativeContainer):
            dependent = providers.Singleton(object)
            singleton = providers.Singleton(Item, dependency=dependent)

        container = Container()

        instance1 = container.singleton()
        with container.reset_singletons():
            instance2 = container.singleton()
        instance3 = container.singleton()

        self.assertEqual(len({instance1, instance2, instance3}), 3)
        self.assertEqual(
            len({instance1.dependency, instance2.dependency, instance3.dependency}),
            3,
        )

    def test_reset_singletons_context_manager_as_attribute(self):
        container = containers.DeclarativeContainer()

        with container.reset_singletons() as alias:
            pass

        self.assertIs(container, alias)

    def test_check_dependencies(self):
        class SubContainer(containers.DeclarativeContainer):
            dependency = providers.Dependency()

        class Container(containers.DeclarativeContainer):
            dependency = providers.Dependency()
            dependencies_container = providers.DependenciesContainer()
            provider = providers.List(dependencies_container.dependency)
            sub_container = providers.Container(SubContainer)

        container = Container()

        with self.assertRaises(errors.Error) as context:
            container.check_dependencies()

        self.assertIn('Container "Container" has undefined dependencies:', str(context.exception))
        self.assertIn('"Container.dependency"', str(context.exception))
        self.assertIn('"Container.dependencies_container.dependency"', str(context.exception))
        self.assertIn('"Container.sub_container.dependency"', str(context.exception))

    def test_check_dependencies_all_defined(self):
        class Container(containers.DeclarativeContainer):
            dependency = providers.Dependency()

        container = Container(dependency='provided')
        result = container.check_dependencies()

        self.assertIsNone(result)

    def test_assign_parent(self):
        parent = providers.DependenciesContainer()
        container = ContainerA()

        container.assign_parent(parent)

        self.assertIs(container.parent, parent)

    def test_parent_name_declarative_parent(self):
        container = ContainerA()
        self.assertEqual(container.parent_name, 'ContainerA')

    def test_parent_name(self):
        container = ContainerA()
        self.assertEqual(container.parent_name, 'ContainerA')

    def test_parent_name_with_deep_parenting(self):
        class Container2(containers.DeclarativeContainer):

            name = providers.Container(ContainerA)

        class Container1(containers.DeclarativeContainer):

            container = providers.Container(Container2)

        container = Container1()
        self.assertEqual(container.container().name.parent_name, 'Container1.container.name')

    def test_parent_name_is_none(self):
        container = containers.DynamicContainer()
        self.assertIsNone(container.parent_name)

    def test_parent_deepcopy(self):
        class Container(containers.DeclarativeContainer):
            container = providers.Container(ContainerA)

        container = Container()

        copied = providers.deepcopy(container)

        self.assertIs(container.container.parent, container)
        self.assertIs(copied.container.parent, copied)

        self.assertIsNot(container, copied)
        self.assertIsNot(container.container, copied.container)
        self.assertIsNot(container.container.parent, copied.container.parent)

    def test_resolve_provider_name(self):
        container = ContainerA()
        self.assertEqual(container.resolve_provider_name(container.p11), 'p11')

    def test_resolve_provider_name_no_provider(self):
        container = ContainerA()
        with self.assertRaises(errors.Error):
            container.resolve_provider_name(providers.Provider())


class SelfTests(unittest.TestCase):

    def test_self(self):
        def call_bar(container):
            return container.bar()

        class Container(containers.DeclarativeContainer):
            __self__ = providers.Self()
            foo = providers.Callable(call_bar, __self__)
            bar = providers.Object('hello')

        container = Container()

        self.assertIs(container.foo(), 'hello')

    def test_self_attribute_implicit(self):
        class Container(containers.DeclarativeContainer):
            pass

        container = Container()

        self.assertIs(container.__self__(), container)

    def test_self_attribute_explicit(self):
        class Container(containers.DeclarativeContainer):
            __self__ = providers.Self()

        container = Container()

        self.assertIs(container.__self__(), container)

    def test_single_self(self):
        with self.assertRaises(errors.Error):
            class Container(containers.DeclarativeContainer):
                self1 = providers.Self()
                self2 = providers.Self()

    def test_self_attribute_alt_name_implicit(self):
        class Container(containers.DeclarativeContainer):
            foo = providers.Self()

        container = Container()

        self.assertIs(container.__self__, container.foo)
        self.assertEqual(set(container.__self__.alt_names), {'foo'})

    def test_self_attribute_alt_name_explicit_1(self):
        class Container(containers.DeclarativeContainer):
            __self__ = providers.Self()
            foo = __self__
            bar = __self__

        container = Container()

        self.assertIs(container.__self__, container.foo)
        self.assertIs(container.__self__, container.bar)
        self.assertEqual(set(container.__self__.alt_names), {'foo', 'bar'})

    def test_self_attribute_alt_name_explicit_2(self):
        class Container(containers.DeclarativeContainer):
            foo = providers.Self()
            bar = foo

        container = Container()

        self.assertIs(container.__self__, container.foo)
        self.assertIs(container.__self__, container.bar)
        self.assertEqual(set(container.__self__.alt_names), {'foo', 'bar'})

    def test_providers_attribute_1(self):
        class Container(containers.DeclarativeContainer):
            __self__ = providers.Self()
            foo = __self__
            bar = __self__

        container = Container()

        self.assertEqual(container.providers, {})
        self.assertEqual(Container.providers, {})

    def test_providers_attribute_2(self):
        class Container(containers.DeclarativeContainer):
            foo = providers.Self()
            bar = foo

        container = Container()

        self.assertEqual(container.providers, {})
        self.assertEqual(Container.providers, {})

    def test_container_multiple_instances(self):
        class Container(containers.DeclarativeContainer):
            __self__ = providers.Self()

        container1 = Container()
        container2 = Container()

        self.assertIsNot(container1, container2)
        self.assertIs(container1.__self__(), container1)
        self.assertIs(container2.__self__(), container2)

    def test_deepcopy(self):
        def call_bar(container):
            return container.bar()

        class Container(containers.DeclarativeContainer):
            __self__ = providers.Self()
            foo = providers.Callable(call_bar, __self__)
            bar = providers.Object('hello')

        container1 = Container()
        container2 = providers.deepcopy(container1)
        container1.bar.override('bye')

        self.assertIs(container1.foo(), 'bye')
        self.assertIs(container2.foo(), 'hello')

    def test_deepcopy_alt_names_1(self):
        class Container(containers.DeclarativeContainer):
            __self__ = providers.Self()
            foo = __self__
            bar = foo

        container1 = Container()
        container2 = providers.deepcopy(container1)

        self.assertIs(container2.__self__(), container2)
        self.assertIs(container2.foo(), container2)
        self.assertIs(container2.bar(), container2)

    def test_deepcopy_alt_names_2(self):
        class Container(containers.DeclarativeContainer):
            self = providers.Self()

        container1 = Container()
        container2 = providers.deepcopy(container1)

        self.assertIs(container2.__self__(), container2)
        self.assertIs(container2.self(), container2)

    def test_deepcopy_no_self_dependencies(self):
        class Container(containers.DeclarativeContainer):
            __self__ = providers.Self()

        container1 = Container()
        container2 = providers.deepcopy(container1)

        self.assertIsNot(container1, container2)
        self.assertIsNot(container1.__self__, container2.__self__)
        self.assertIs(container1.__self__(), container1)
        self.assertIs(container2.__self__(), container2)

    def test_with_container_provider(self):
        def call_bar(container):
            return container.bar()

        class SubContainer(containers.DeclarativeContainer):
            __self__ = providers.Self()
            foo = providers.Callable(call_bar, __self__)
            bar = providers.Object('hello')

        class Container(containers.DeclarativeContainer):
            sub_container = providers.Container(SubContainer)

            baz = providers.Callable(lambda value: value, sub_container.foo)

        container = Container()

        self.assertIs(container.baz(), 'hello')

    def test_with_container_provider_overriding(self):
        def call_bar(container):
            return container.bar()

        class SubContainer(containers.DeclarativeContainer):
            __self__ = providers.Self()
            foo = providers.Callable(call_bar, __self__)
            bar = providers.Object('hello')

        class Container(containers.DeclarativeContainer):
            sub_container = providers.Container(SubContainer, bar='bye')

            baz = providers.Callable(lambda value: value, sub_container.foo)

        container = Container()

        self.assertIs(container.baz(), 'bye')

    def test_with_container_provider_self(self):
        class SubContainer(containers.DeclarativeContainer):
            __self__ = providers.Self()

        class Container(containers.DeclarativeContainer):
            sub_container = providers.Container(SubContainer)

        container = Container()

        self.assertIs(container.__self__(), container)
        self.assertIs(container.sub_container().__self__(), container.sub_container())


class DynamicContainerWithCustomStringTests(unittest.TestCase):
    # See: https://github.com/ets-labs/python-dependency-injector/issues/479

    class CustomString(str):
        pass

    class CustomClass:
        thing = None

    def setUp(self):
        self.container = containers.DynamicContainer()
        self.provider = providers.Provider()

    def test_setattr(self):
        setattr(self.container, self.CustomString('test_attr'), self.provider)
        self.assertIs(self.container.test_attr, self.provider)

    def test_delattr(self):
        setattr(self.container, self.CustomString('test_attr'), self.provider)
        delattr(self.container, self.CustomString('test_attr'))
        with self.assertRaises(AttributeError):
            self.container.test_attr

    def test_set_provider(self):
        self.container.set_provider(self.CustomString('test_attr'), self.provider)
        self.assertIs(self.container.test_attr, self.provider)

    def test_set_providers(self):
        self.container.set_providers(**{self.CustomString('test_attr'): self.provider})
        self.assertIs(self.container.test_attr, self.provider)



import unittest

from dependency_injector import containers, providers


class TraverseProviderTests(unittest.TestCase):

    def test_nested_providers(self):
        class Container(containers.DeclarativeContainer):
            obj_factory = providers.DelegatedFactory(
                dict,
                foo=providers.Resource(
                    dict,
                    foo='bar'
                ),
                bar=providers.Resource(
                    dict,
                    foo='bar'
                )
            )

        container = Container()
        all_providers = list(container.traverse())

        self.assertIn(container.obj_factory, all_providers)
        self.assertIn(container.obj_factory.kwargs['foo'], all_providers)
        self.assertIn(container.obj_factory.kwargs['bar'], all_providers)
        self.assertEqual(len(all_providers), 3)

    def test_nested_providers_with_filtering(self):
        class Container(containers.DeclarativeContainer):
            obj_factory = providers.DelegatedFactory(
                dict,
                foo=providers.Resource(
                    dict,
                    foo='bar'
                ),
                bar=providers.Resource(
                    dict,
                    foo='bar'
                )
            )

        container = Container()
        all_providers = list(container.traverse(types=[providers.Resource]))

        self.assertIn(container.obj_factory.kwargs['foo'], all_providers)
        self.assertIn(container.obj_factory.kwargs['bar'], all_providers)
        self.assertEqual(len(all_providers), 2)


class TraverseProviderDeclarativeTests(unittest.TestCase):

    def test_nested_providers(self):
        class Container(containers.DeclarativeContainer):
            obj_factory = providers.DelegatedFactory(
                dict,
                foo=providers.Resource(
                    dict,
                    foo='bar'
                ),
                bar=providers.Resource(
                    dict,
                    foo='bar'
                )
            )

        all_providers = list(Container.traverse())

        self.assertIn(Container.obj_factory, all_providers)
        self.assertIn(Container.obj_factory.kwargs['foo'], all_providers)
        self.assertIn(Container.obj_factory.kwargs['bar'], all_providers)
        self.assertEqual(len(all_providers), 3)

    def test_nested_providers_with_filtering(self):
        class Container(containers.DeclarativeContainer):
            obj_factory = providers.DelegatedFactory(
                dict,
                foo=providers.Resource(
                    dict,
                    foo='bar'
                ),
                bar=providers.Resource(
                    dict,
                    foo='bar'
                )
            )

        all_providers = list(Container.traverse(types=[providers.Resource]))

        self.assertIn(Container.obj_factory.kwargs['foo'], all_providers)
        self.assertIn(Container.obj_factory.kwargs['bar'], all_providers)
        self.assertEqual(len(all_providers), 2)



"""Dependency injector resource provider unit tests."""

import asyncio

import unittest

from dependency_injector import containers, providers, resources, errors

# Runtime import to get asyncutils module
import os
_TOP_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        '../',
    )),
)
import sys
sys.path.append(_TOP_DIR)

from asyncutils import AsyncTestCase


def init_fn(*args, **kwargs):
    return args, kwargs


class ResourceTests(unittest.TestCase):

    def test_is_provider(self):
        self.assertTrue(providers.is_provider(providers.Resource(init_fn)))

    def test_init_optional_provides(self):
        provider = providers.Resource()
        provider.set_provides(init_fn)
        self.assertIs(provider.provides, init_fn)
        self.assertEqual(provider(), (tuple(), dict()))

    def test_set_provides_returns_self(self):
        provider = providers.Resource()
        self.assertIs(provider.set_provides(init_fn), provider)

    def test_provided_instance_provider(self):
        provider = providers.Resource(init_fn)
        self.assertIsInstance(provider.provided, providers.ProvidedInstance)

    def test_injection(self):
        resource = object()

        def _init():
            _init.counter += 1
            return resource
        _init.counter = 0

        class Container(containers.DeclarativeContainer):
            resource = providers.Resource(_init)
            dependency1 = providers.List(resource)
            dependency2 = providers.List(resource)

        container = Container()
        list1 = container.dependency1()
        list2 = container.dependency2()

        self.assertEqual(list1, [resource])
        self.assertIs(list1[0], resource)

        self.assertEqual(list2, [resource])
        self.assertIs(list2[0], resource)

        self.assertEqual(_init.counter, 1)

    def test_init_function(self):
        def _init():
            _init.counter += 1
        _init.counter = 0

        provider = providers.Resource(_init)

        result1 = provider()
        self.assertIsNone(result1)
        self.assertEqual(_init.counter, 1)

        result2 = provider()
        self.assertIsNone(result2)
        self.assertEqual(_init.counter, 1)

        provider.shutdown()

    def test_init_generator(self):
        def _init():
            _init.init_counter += 1
            yield
            _init.shutdown_counter += 1

        _init.init_counter = 0
        _init.shutdown_counter = 0

        provider = providers.Resource(_init)

        result1 = provider()
        self.assertIsNone(result1)
        self.assertEqual(_init.init_counter, 1)
        self.assertEqual(_init.shutdown_counter, 0)

        provider.shutdown()
        self.assertEqual(_init.init_counter, 1)
        self.assertEqual(_init.shutdown_counter, 1)

        result2 = provider()
        self.assertIsNone(result2)
        self.assertEqual(_init.init_counter, 2)
        self.assertEqual(_init.shutdown_counter, 1)

        provider.shutdown()
        self.assertEqual(_init.init_counter, 2)
        self.assertEqual(_init.shutdown_counter, 2)

    def test_init_class(self):
        class TestResource(resources.Resource):
            init_counter = 0
            shutdown_counter = 0

            def init(self):
                self.__class__.init_counter += 1

            def shutdown(self, _):
                self.__class__.shutdown_counter += 1

        provider = providers.Resource(TestResource)

        result1 = provider()
        self.assertIsNone(result1)
        self.assertEqual(TestResource.init_counter, 1)
        self.assertEqual(TestResource.shutdown_counter, 0)

        provider.shutdown()
        self.assertEqual(TestResource.init_counter, 1)
        self.assertEqual(TestResource.shutdown_counter, 1)

        result2 = provider()
        self.assertIsNone(result2)
        self.assertEqual(TestResource.init_counter, 2)
        self.assertEqual(TestResource.shutdown_counter, 1)

        provider.shutdown()
        self.assertEqual(TestResource.init_counter, 2)
        self.assertEqual(TestResource.shutdown_counter, 2)

    def test_init_not_callable(self):
        provider = providers.Resource(1)
        with self.assertRaises(errors.Error):
            provider.init()

    def test_init_and_shutdown(self):
        def _init():
            _init.init_counter += 1
            yield
            _init.shutdown_counter += 1

        _init.init_counter = 0
        _init.shutdown_counter = 0

        provider = providers.Resource(_init)

        result1 = provider.init()
        self.assertIsNone(result1)
        self.assertEqual(_init.init_counter, 1)
        self.assertEqual(_init.shutdown_counter, 0)

        provider.shutdown()
        self.assertEqual(_init.init_counter, 1)
        self.assertEqual(_init.shutdown_counter, 1)

        result2 = provider.init()
        self.assertIsNone(result2)
        self.assertEqual(_init.init_counter, 2)
        self.assertEqual(_init.shutdown_counter, 1)

        provider.shutdown()
        self.assertEqual(_init.init_counter, 2)
        self.assertEqual(_init.shutdown_counter, 2)

    def test_shutdown_of_not_initialized(self):
        def _init():
            yield

        provider = providers.Resource(_init)

        result = provider.shutdown()
        self.assertIsNone(result)

    def test_initialized(self):
        provider = providers.Resource(init_fn)
        self.assertFalse(provider.initialized)

        provider.init()
        self.assertTrue(provider.initialized)

        provider.shutdown()
        self.assertFalse(provider.initialized)

    def test_call_with_context_args(self):
        provider = providers.Resource(init_fn, 'i1', 'i2')
        self.assertEqual(provider('i3', i4=4), (('i1', 'i2', 'i3'), {'i4': 4}))

    def test_fluent_interface(self):
        provider = providers.Resource(init_fn) \
            .add_args(1, 2) \
            .add_kwargs(a3=3, a4=4)

        self.assertEqual(provider(), ((1, 2), {'a3': 3, 'a4': 4}))

    def test_set_args(self):
        provider = providers.Resource(init_fn) \
            .add_args(1, 2) \
            .set_args(3, 4)
        self.assertEqual(provider.args, (3, 4))

    def test_clear_args(self):
        provider = providers.Resource(init_fn) \
            .add_args(1, 2) \
            .clear_args()
        self.assertEqual(provider.args, tuple())

    def test_set_kwargs(self):
        provider = providers.Resource(init_fn) \
            .add_kwargs(a1='i1', a2='i2') \
            .set_kwargs(a3='i3', a4='i4')
        self.assertEqual(provider.kwargs, {'a3': 'i3', 'a4': 'i4'})

    def test_clear_kwargs(self):
        provider = providers.Resource(init_fn) \
            .add_kwargs(a1='i1', a2='i2') \
            .clear_kwargs()
        self.assertEqual(provider.kwargs, {})

    def test_call_overridden(self):
        provider = providers.Resource(init_fn, 1)
        overriding_provider1 = providers.Resource(init_fn, 2)
        overriding_provider2 = providers.Resource(init_fn, 3)

        provider.override(overriding_provider1)
        provider.override(overriding_provider2)

        instance1 = provider()
        instance2 = provider()

        self.assertIs(instance1, instance2)
        self.assertEqual(instance1, ((3,), {}))
        self.assertEqual(instance2, ((3,), {}))

    def test_deepcopy(self):
        provider = providers.Resource(init_fn, 1, 2, a3=3, a4=4)

        provider_copy = providers.deepcopy(provider)

        self.assertIsNot(provider, provider_copy)
        self.assertEqual(provider.args, provider_copy.args)
        self.assertEqual(provider.kwargs, provider_copy.kwargs)
        self.assertIsInstance(provider, providers.Resource)

    def test_deepcopy_initialized(self):
        provider = providers.Resource(init_fn)
        provider.init()

        with self.assertRaises(errors.Error):
            providers.deepcopy(provider)

    def test_deepcopy_from_memo(self):
        provider = providers.Resource(init_fn)
        provider_copy_memo = providers.Resource(init_fn)

        provider_copy = providers.deepcopy(
            provider,
            memo={id(provider): provider_copy_memo},
        )

        self.assertIs(provider_copy, provider_copy_memo)

    def test_deepcopy_args(self):
        provider = providers.Resource(init_fn)
        dependent_provider1 = providers.Factory(list)
        dependent_provider2 = providers.Factory(dict)

        provider.add_args(dependent_provider1, dependent_provider2)

        provider_copy = providers.deepcopy(provider)
        dependent_provider_copy1 = provider_copy.args[0]
        dependent_provider_copy2 = provider_copy.args[1]

        self.assertNotEqual(provider.args, provider_copy.args)

        self.assertIs(dependent_provider1.cls, dependent_provider_copy1.cls)
        self.assertIsNot(dependent_provider1, dependent_provider_copy1)

        self.assertIs(dependent_provider2.cls, dependent_provider_copy2.cls)
        self.assertIsNot(dependent_provider2, dependent_provider_copy2)

    def test_deepcopy_kwargs(self):
        provider = providers.Resource(init_fn)
        dependent_provider1 = providers.Factory(list)
        dependent_provider2 = providers.Factory(dict)

        provider.add_kwargs(d1=dependent_provider1, d2=dependent_provider2)

        provider_copy = providers.deepcopy(provider)
        dependent_provider_copy1 = provider_copy.kwargs['d1']
        dependent_provider_copy2 = provider_copy.kwargs['d2']

        self.assertNotEqual(provider.kwargs, provider_copy.kwargs)

        self.assertIs(dependent_provider1.cls, dependent_provider_copy1.cls)
        self.assertIsNot(dependent_provider1, dependent_provider_copy1)

        self.assertIs(dependent_provider2.cls, dependent_provider_copy2.cls)
        self.assertIsNot(dependent_provider2, dependent_provider_copy2)

    def test_deepcopy_overridden(self):
        provider = providers.Resource(init_fn)
        object_provider = providers.Object(object())

        provider.override(object_provider)

        provider_copy = providers.deepcopy(provider)
        object_provider_copy = provider_copy.overridden[0]

        self.assertIsNot(provider, provider_copy)
        self.assertEqual(provider.args, provider_copy.args)
        self.assertIsInstance(provider, providers.Resource)

        self.assertIsNot(object_provider, object_provider_copy)
        self.assertIsInstance(object_provider_copy, providers.Object)

    def test_deepcopy_with_sys_streams(self):
        provider = providers.Resource(init_fn)
        provider.add_args(sys.stdin, sys.stdout, sys.stderr)

        provider_copy = providers.deepcopy(provider)

        self.assertIsNot(provider, provider_copy)
        self.assertIsInstance(provider_copy, providers.Resource)
        self.assertIs(provider.args[0], sys.stdin)
        self.assertIs(provider.args[1], sys.stdout)
        self.assertIs(provider.args[2], sys.stderr)

    def test_repr(self):
        provider = providers.Resource(init_fn)

        self.assertEqual(
            repr(provider),
            '<dependency_injector.providers.Resource({0}) at {1}>'.format(
                repr(init_fn),
                hex(id(provider)),
            )
        )


class AsyncResourceTest(AsyncTestCase):

    def test_init_async_function(self):
        resource = object()

        async def _init():
            await asyncio.sleep(0.001)
            _init.counter += 1
            return resource
        _init.counter = 0

        provider = providers.Resource(_init)

        result1 = self._run(provider())
        self.assertIs(result1, resource)
        self.assertEqual(_init.counter, 1)

        result2 = self._run(provider())
        self.assertIs(result2, resource)
        self.assertEqual(_init.counter, 1)

        self._run(provider.shutdown())

    def test_init_async_generator(self):
        resource = object()

        async def _init():
            await asyncio.sleep(0.001)
            _init.init_counter += 1

            yield resource

            await asyncio.sleep(0.001)
            _init.shutdown_counter += 1

        _init.init_counter = 0
        _init.shutdown_counter = 0

        provider = providers.Resource(_init)

        result1 = self._run(provider())
        self.assertIs(result1, resource)
        self.assertEqual(_init.init_counter, 1)
        self.assertEqual(_init.shutdown_counter, 0)

        self._run(provider.shutdown())
        self.assertEqual(_init.init_counter, 1)
        self.assertEqual(_init.shutdown_counter, 1)

        result2 = self._run(provider())
        self.assertIs(result2, resource)
        self.assertEqual(_init.init_counter, 2)
        self.assertEqual(_init.shutdown_counter, 1)

        self._run(provider.shutdown())
        self.assertEqual(_init.init_counter, 2)
        self.assertEqual(_init.shutdown_counter, 2)

    def test_init_async_class(self):
        resource = object()

        class TestResource(resources.AsyncResource):
            init_counter = 0
            shutdown_counter = 0

            async def init(self):
                await asyncio.sleep(0.001)
                self.__class__.init_counter += 1
                return resource

            async def shutdown(self, resource_):
                await asyncio.sleep(0.001)
                self.__class__.shutdown_counter += 1
                assert resource_ is resource

        provider = providers.Resource(TestResource)

        result1 = self._run(provider())
        self.assertIs(result1, resource)
        self.assertEqual(TestResource.init_counter, 1)
        self.assertEqual(TestResource.shutdown_counter, 0)

        self._run(provider.shutdown())
        self.assertEqual(TestResource.init_counter, 1)
        self.assertEqual(TestResource.shutdown_counter, 1)

        result2 = self._run(provider())
        self.assertIs(result2, resource)
        self.assertEqual(TestResource.init_counter, 2)
        self.assertEqual(TestResource.shutdown_counter, 1)

        self._run(provider.shutdown())
        self.assertEqual(TestResource.init_counter, 2)
        self.assertEqual(TestResource.shutdown_counter, 2)

    def test_init_with_error(self):
        async def _init():
            raise RuntimeError()

        provider = providers.Resource(_init)

        future = provider()
        self.assertTrue(provider.initialized)
        self.assertTrue(provider.is_async_mode_enabled())

        with self.assertRaises(RuntimeError):
            self._run(future)

        self.assertFalse(provider.initialized)
        self.assertTrue(provider.is_async_mode_enabled())

    def test_init_async_gen_with_error(self):
        async def _init():
            raise RuntimeError()
            yield

        provider = providers.Resource(_init)

        future = provider()
        self.assertTrue(provider.initialized)
        self.assertTrue(provider.is_async_mode_enabled())

        with self.assertRaises(RuntimeError):
            self._run(future)

        self.assertFalse(provider.initialized)
        self.assertTrue(provider.is_async_mode_enabled())

    def test_init_async_subclass_with_error(self):
        class _Resource(resources.AsyncResource):
            async def init(self):
                raise RuntimeError()

            async def shutdown(self, resource):
                pass

        provider = providers.Resource(_Resource)

        future = provider()
        self.assertTrue(provider.initialized)
        self.assertTrue(provider.is_async_mode_enabled())

        with self.assertRaises(RuntimeError):
            self._run(future)

        self.assertFalse(provider.initialized)
        self.assertTrue(provider.is_async_mode_enabled())

    def test_init_with_dependency_to_other_resource(self):
        # See: https://github.com/ets-labs/python-dependency-injector/issues/361
        async def init_db_connection(db_url: str):
            await asyncio.sleep(0.001)
            yield {'connection': 'ok', 'url': db_url}

        async def init_user_session(db):
            await asyncio.sleep(0.001)
            yield {'session': 'ok', 'db': db}

        class Container(containers.DeclarativeContainer):
            config = providers.Configuration()

            db_connection = providers.Resource(
                init_db_connection,
                db_url=config.db_url,
            )

            user_session = providers.Resource(
                init_user_session,
                db=db_connection
            )

        async def main():
            container = Container(config={'db_url': 'postgres://...'})
            try:
                return await container.user_session()
            finally:
                await container.shutdown_resources()

        result = self._run(main())

        self.assertEqual(
            result,
            {'session': 'ok', 'db': {'connection': 'ok', 'url': 'postgres://...'}},
        )

    def test_init_and_shutdown_methods(self):
        async def _init():
            await asyncio.sleep(0.001)
            _init.init_counter += 1

            yield

            await asyncio.sleep(0.001)
            _init.shutdown_counter += 1

        _init.init_counter = 0
        _init.shutdown_counter = 0

        provider = providers.Resource(_init)

        self._run(provider.init())
        self.assertEqual(_init.init_counter, 1)
        self.assertEqual(_init.shutdown_counter, 0)

        self._run(provider.shutdown())
        self.assertEqual(_init.init_counter, 1)
        self.assertEqual(_init.shutdown_counter, 1)

        self._run(provider.init())
        self.assertEqual(_init.init_counter, 2)
        self.assertEqual(_init.shutdown_counter, 1)

        self._run(provider.shutdown())
        self.assertEqual(_init.init_counter, 2)
        self.assertEqual(_init.shutdown_counter, 2)

    def test_shutdown_of_not_initialized(self):
        async def _init():
            yield

        provider = providers.Resource(_init)
        provider.enable_async_mode()

        result = self._run(provider.shutdown())
        self.assertIsNone(result)

    def test_concurrent_init(self):
        resource = object()

        async def _init():
            await asyncio.sleep(0.001)
            _init.counter += 1
            return resource
        _init.counter = 0

        provider = providers.Resource(_init)

        result1, result2 = self._run(
            asyncio.gather(
                provider(),
                provider()
            ),
        )

        self.assertIs(result1, resource)
        self.assertEqual(_init.counter, 1)

        self.assertIs(result2, resource)
        self.assertEqual(_init.counter, 1)
